from stx.design.document import Document
from stx.loading.linking import link_component
from stx.loading.numbering import build_numbering
from stx.parsing import parse_from_file


def from_file(file_path: str) -> Document:
    doc = Document()

    doc.content = parse_from_file(doc, file_path)

    doc.refs = link_component(doc.content)

    doc.index = build_numbering(doc.content)

    return doc



from typing import List

from stx import logger
from stx.design.index_node import IndexNode
from stx.design.components import Component, Composite, CodeBlock, Heading, Table, \
    ListBlock, TextBlock, RawText, PlainText, StyledText, LinkText, Figure
from stx.design.document import Document

from stx.rendering.html5.writer import HtmlWriter


def render_document(document: Document, writer: HtmlWriter):
    logger.info('Rendering HTML document...')

    writer.write('<!DOCTYPE html>\n')
    writer.open_tag('html')

    writer.open_tag('head')

    writer.tag('meta', {'author': 'generator', 'content': 'Sergio Pedraza'})
    writer.tag('meta', {'name': 'generator', 'content': 'STX 0.0.1'})

    writer.open_tag('title')

    # TODO improve this
    writer.text(document.index[0].heading.get_text())

    writer.close_tag('title')

    for stylesheet in document.links.get_list('stylesheet'):
        writer.tag('link', {
            'rel': 'stylesheet',
            'type': 'text/css',
            'href': stylesheet,
        })

    writer.close_tag('head')

    writer.open_tag('body')

    render_index(document, writer)

    writer.open_tag('main')

    render_content(document, writer, document.content)

    writer.close_tag('main')

    writer.close_tag('body')

    writer.close_tag('html')


def render_container(document: Document, writer: HtmlWriter, container: Composite):
    render_contents(document, writer, container.components)


def render_contents(document: Document, writer: HtmlWriter, contents: List[Component]):
    for content in contents:
        render_content(document, writer, content)


def render_code_block(document: Document, writer: HtmlWriter, code: CodeBlock):
    pre_attributes = {}

    css_classes = code.attributes.get_list('source')

    if len(css_classes) > 0:
        pre_attributes['class'] = ' '.join(css_classes)

    writer.open_tag('pre', inline=True, attributes=pre_attributes)
    writer.text(code.code, disable_indentation=True)
    writer.close_tag('pre', inline=True)
    writer.break_line()


def open_tag(
        document: Document,
        writer: HtmlWriter,
        content: Component,
        tag: str,
        attrs: dict):
    main_ref = document.refs.get_main_ref(content)

    if main_ref is not None:
        attrs['id'] = main_ref

    writer.open_tag(tag, attrs)

    for other_ref in document.refs.get_other_refs(content):
        writer.open_tag('a', {'id': other_ref}, inline=True)
        writer.close_tag('a', inline=True)


def render_heading(document: Document, writer: HtmlWriter, heading: Heading):
    level = heading.level

    if level < 1:
        level = 1
    elif level > 6:
        level = 6

    tag = f'h{level}'
    attrs = {}

    open_tag(document, writer, heading, tag, attrs)

    if heading.number is not None:
        writer.open_tag('span', inline=True)
        writer.text(heading.number)
        writer.text(' ')
        writer.close_tag('span', inline=True)

    render_content(document, writer, heading.content, collapse_paragraph=True)

    writer.close_tag(tag)


def render_table(
        document: Document,
        writer: HtmlWriter,
        table: Table):
    writer.open_tag('table')

    if table.caption is not None:
        writer.open_tag('caption')

        if table.number is not None:
            writer.open_tag('span', {'class': 'stx-number'}, inline=True)
            writer.text(table.number)
            writer.close_tag('span', inline=True)
            writer.text(' ')

        render_content(document, writer, table.caption, collapse_paragraph=True)

        writer.close_tag('caption')

    for row in table.rows:
        writer.open_tag('tr')

        if row.header:
            cell_tag = 'th'
        else:
            cell_tag = 'td'

        for cell in row.cells:
            writer.open_tag(cell_tag)

            render_content(document, writer, cell, collapse_paragraph=True)

            writer.close_tag(cell_tag)

        writer.close_tag('tr')

    writer.close_tag('table')


def render_list(document: Document, writer: HtmlWriter, lst: ListBlock):
    if lst.ordered:
        tag = 'ol'
    else:
        tag = 'ul'

    writer.open_tag(tag)

    for item in lst.items:
        writer.open_tag('li')

        render_content(document, writer, item, collapse_paragraph=True)

        writer.close_tag('li')

    writer.close_tag(tag)


def render_paragraph(document: Document, writer: HtmlWriter, paragraph: TextBlock):
    open_tag(document, writer, paragraph, 'p', {})

    for content in paragraph.components:
        render_content(document, writer, content)

    writer.close_tag('p')


def render_raw(document: Document, writer: HtmlWriter, raw: RawText):
    writer.text(raw.text)  # TODO to delete?


def render_plain_text(document: Document, writer: HtmlWriter, plain: PlainText):
    writer.text(plain.text)


def render_styled_text(document: Document, writer: HtmlWriter, styled: StyledText):
    if styled.style == 'strong':
        tag = 'strong'
    elif styled.style == 'emphasized':
        tag = 'em'
    elif styled.style == 'code':
        tag = 'code'
    else:
        tag = 'span'

    writer.open_tag(tag, inline=True)

    render_contents(document, writer, styled.contents)

    writer.close_tag(tag, inline=True)


def render_link_text(document: Document, writer: HtmlWriter, link: LinkText):
    if link.internal:
        href = f'#{link.reference}'
    else:
        href = link.reference

    writer.open_tag('a', {
        'href': href,
    }, inline=True)

    render_contents(document, writer, link.contents)

    writer.close_tag('a', inline=True)


def render_embedded_text(
        document: Document, writer: HtmlWriter, embedded: RawText):
    # writer.comment(embedded.source) TODO where is comes?
    writer.write_raw(embedded.text)


def render_figure(document: Document, writer: HtmlWriter, figure: Figure):
    writer.open_tag('figure')

    render_content(document, writer, figure.content)

    writer.open_tag('figcaption')

    if figure.number is not None:
        writer.open_tag('span', {'class': 'stx-number'}, inline=True)
        writer.text(figure.number)
        writer.close_tag('span', inline=True)
        writer.text(' ')

    render_content(document, writer, figure.caption, collapse_paragraph=True)

    writer.close_tag('figcaption')

    writer.close_tag('figure')


def render_content(
        document: Document,
        writer: HtmlWriter,
        content: Component,
        collapse_paragraph=False):
    if collapse_paragraph and isinstance(content, TextBlock):
        render_contents(document, writer, content.components)
    elif isinstance(content, Composite):
        render_container(document, writer, content)
    elif isinstance(content, CodeBlock):
        render_code_block(document, writer, content)
    elif isinstance(content, Heading):
        render_heading(document, writer, content)
    elif isinstance(content, Table):
        render_table(document, writer, content)
    elif isinstance(content, ListBlock):
        render_list(document, writer, content)
    elif isinstance(content, TextBlock):
        render_paragraph(document, writer, content)
    elif isinstance(content, PlainText):
        render_plain_text(document, writer, content)
    elif isinstance(content, StyledText):
        render_styled_text(document, writer, content)
    elif isinstance(content, LinkText):
        render_link_text(document, writer, content)
    elif isinstance(content, RawText):
        render_embedded_text(document, writer, content)
    elif isinstance(content, Figure):
        render_figure(document, writer, content)
    else:
        raise NotImplementedError()


def render_index(document: Document, writer: HtmlWriter):
    writer.open_tag('nav', {'class': 'toc'})

    render_index_nodes(document, writer, document.index)

    writer.close_tag('nav')


def render_index_nodes(
        document: Document, writer: HtmlWriter, nodes: List[IndexNode]):
    if len(nodes) == 0:
        return

    writer.open_tag('ol')

    for node in nodes:
        ref = document.refs.get_main_ref(node.heading)

        a_attrs = {}

        if ref is not None:
            a_attrs['href'] = f'#{ref}'

        writer.open_tag('li')

        writer.open_tag('a', a_attrs, inline=True)

        if node.heading.number is not None:
            writer.text(node.heading.number)
            writer.text(' ')

        render_content(
            document, writer, node.heading.content, collapse_paragraph=True)

        writer.close_tag('a', inline=True)
        writer.break_line()

        render_index_nodes(document, writer, node.nodes)

        writer.close_tag('li')

    writer.close_tag('ol')



import click

from stx.loading.loaders import from_file
from stx.rendering.html5.renderer import render_document
from stx.rendering.html5.writer import HtmlWriter


@click.command()
@click.argument('input_file')
@click.argument('output_file')
@click.option('--output_format', default='html5')
@click.option('--output_encoding', default='utf-8')
def main(
        input_file: str,
        output_file: str,
        output_format: str,
        output_encoding: str):
    document = from_file(input_file)

    if output_format == 'html5':
        with open(output_file, 'w', encoding=output_encoding) as f:
            render_document(document, HtmlWriter(f))
    else:
        raise Exception(f'Not implemented output format: {output_format}')


if __name__ == '__main__':
    main()



from stx import logger
from stx.design.components import Component, Heading, LinkText
from stx.design.ref_map import make_ref, RefMap


def link_component(component: Component) -> RefMap:
    ref_map = RefMap()

    logger.info('Processing `ref` attributes...')
    consume_ref_attribute(ref_map, component)

    logger.info('Building automatic references...')
    build_content(ref_map, component)

    logger.info('Validating links references...')
    validate_content(ref_map, component)

    return ref_map


def consume_ref_attribute(ref_map: RefMap, component: Component):
    for child in component.walk():
        for ref in child.attributes.get_list('ref'):
            ref = make_ref(ref)

            ref_map.register_ref(ref, child)


def build_heading(ref_map: RefMap, heading: Heading):
    ref_map.register_content(heading, main=True)


def build_content(ref_map: RefMap, content: Component):
    if isinstance(content, Heading):
        build_heading(ref_map, content)

    for child in content.get_children():
        build_content(ref_map, child)


def validate_link(ref_map: RefMap, link: LinkText):
    if link.reference is not None and not link.internal:
        return

    if link.reference is not None:
        reference = link.reference
    else:
        reference = link.get_text()

    ref = make_ref(reference)

    if not ref_map.contains_ref(ref):
        print(f'WARNING: Invalid reference: {reference}')

    link.reference = ref


def validate_content(ref_map: RefMap, content: Component):
    for child in content.walk():
        if isinstance(child, LinkText):
            validate_link(ref_map, child)



from __future__ import annotations

from typing import Optional, List

from stx.design.attributes_map import AttributesMap
from stx.design.components import Component, Table, Figure


class Composer:

    def __init__(self):
        self.attributes = AttributesMap()
        self.contents: List[Optional[Component]] = []
        self.pending_caption: Optional[Component] = None

    def push(self, content: Component):
        if self.pending_caption is not None:
            if isinstance(content, Table):
                content.caption = self.pending_caption
            else:
                content = Figure(content, self.pending_caption)

            self.pending_caption = None

        # Take buffered attributes
        content.attributes.update(self.attributes)
        self.attributes.clear()

        self.contents.append(content)

    def pop(self) -> Component:
        if len(self.contents) == 0:
            raise Exception('Expected content')

        return self.contents.pop()

    def compile(self) -> List[Component]:
        if self.pending_caption is not None:
            raise Exception('pending cation')

        return [c for c in self.contents if c is not None]

    def get_last_content(self) -> Optional[Component]:
        if len(self.contents) == 0:
            return None

        return self.contents[-1]

    def push_separator(self):  # TODO still valid?
        self.contents.append(None)

    def push_attribute(self, name: str, values: List[str]):
        self.attributes.add_values(name, values)





from stx.design.attributes_map import AttributesMap


class Document:

    def __init__(self):
        self.content = None
        self.refs = None
        self.index = None
        self.links = AttributesMap()



from __future__ import annotations

from typing import Optional, TextIO, List

from stx.errors import ParseError

from . import _marks


class Source:

    def __init__(self, stream: TextIO, file_path: str):
        self._stream = stream
        self.file_path = file_path
        self._begin_index_stack: List[int] = []
        self._curr_line: Optional[Line] = None
        self._next_index = 0

    def try_line(self) -> Optional[Line]:
        if self._curr_line is None:
            if self._stream is None:
                return None

            text = self._stream.readline()

            if text == '':
                self._stream = None
                return None

            text = text.rstrip('\r\n')

            self._curr_line = Line(self, text, self._next_index)
            self._next_index += 1

        if self._curr_line.indentation < self.peek_begin_index() and not self._curr_line.is_empty:
            return None

        return self._curr_line

    def is_alive(self) -> bool:
        return self.try_line() is not None

    def get_line(self) -> Line:
        line = self.try_line()

        if line is None:
            raise self.error('Expected to read a line.')

        return line

    def pop_line(self) -> Line:
        line = self.get_line()

        self._curr_line = None

        return line

    def flush_line(self):
        self._curr_line = None

    def peek_begin_index(self) -> int:
        if len(self._begin_index_stack) == 0:
            return 0

        return self._begin_index_stack[-1]

    def push_begin_index(self, begin_index: int):
        self._begin_index_stack.append(begin_index)

    def pop_begin_index(self):
        self._begin_index_stack.pop()

    def error(self, message: str) -> ParseError:
        return ParseError(
            message=message,
            file_path=self.file_path,
            line_index=self._next_index - 1,
            column_index=self.peek_begin_index(),
        )


class Line:

    def __init__(self, source: Source, raw_text: str, index: int):
        self.source = source
        self.raw_text = raw_text
        self.index = index

        raw_length = len(raw_text)
        indentation = 0

        for c in raw_text:
            if c == ' ':
                indentation += 1
            else:
                break

        self.raw_length = raw_length
        self.is_empty = (indentation == raw_length)
        self.indentation = indentation

    def __repr__(self):
        return f'Line {self.index+1}: {self.raw_text}'

    def get_current_text(self) -> str:
        begin_index = self.source.peek_begin_index()

        if begin_index < self.indentation:
            begin_index = self.indentation

        return self.raw_text[begin_index:]

    def test_current_text(self, expected: str, strip=True) -> bool:
        actual = self.get_current_text()

        if strip:
            actual = actual.strip()

        return actual == expected

    def consume_block_info(self) -> Optional[str]:
        text = self.get_current_text()

        for mark in _marks.reserved_block_marks:
            if text.startswith(mark):
                begin_index = self.indentation + len(mark)

                while begin_index < self.raw_length:
                    if self.raw_text[begin_index] == ' ':
                        begin_index += 1
                    else:
                        break

                self.indentation = begin_index

                return mark

        return None

    def error(self, message: str, column_index: int) -> ParseError:
        return ParseError(
            message=message,
            file_path=self.source.file_path,
            line_index=self.index,
            column_index=column_index,
        )



from __future__ import annotations

from typing import Optional

from stx.errors import ParseError

from ._source import Source


class Tape:

    def __init__(self, source: Source):
        self.source = source
        self.text = ''
        self.position = 0
        self.line_index = None
        self.column_index = None

    def skip(self, targets: str) -> int:
        count = 0

        while True:
            c = self.try_peek()

            if c is None or c not in targets:
                break

            count += 1
            self.position += 1
            self.column_index += 1

        return count

    def try_peek(self) -> Optional[str]:
        if self.text is None:
            return None
        elif self.position >= len(self.text):
            line = self.source.try_line()

            if line is None:
                self.text = None
                return None

            self.source.flush_line()

            self.line_index = line.index
            self.column_index = 0

            if line.is_empty:
                self.text = None
                return None

            self.text += line.get_current_text() + '\n'

        return self.text[self.position]

    def peek(self) -> str:
        c = self.try_peek()

        if c is None:
            raise self.error('Expected to read a char.')

        return c

    def read(self) -> str:
        c = self.peek()

        self.position += 1
        self.column_index += 1

        return c

    def alive(self) -> bool:
        return self.try_peek() is not None

    def test(self, expected: str) -> bool:
        if expected is None:
            return False

        actual = self.try_peek()

        if actual is None:
            return False

        return actual in expected

    def any(self, *chars: str) -> bool:
        for c in chars:
            if self.test(c):
                return True
        return False

    def pull(self, c: str) -> bool:
        if self.test(c):
            self.read()
            return True

        return False

    def error(self, message: str) -> ParseError:
        return ParseError(
            message=message,
            file_path=self.source.file_path,
            line_index=self.line_index,
            column_index=self.column_index,
        )




from __future__ import annotations

from typing import Dict, Optional, List


class AttributesMap:

    def __init__(self):
        self._data: Dict[str, List[str]] = {}
        self._read: Dict[str, bool] = {}

    def update(self, other_map: AttributesMap):
        self._data = dict(other_map._data)
        self._read.clear()

    def clear(self):
        self._data.clear()
        self._read.clear()

    def add_values(self, name: str, values: List[str]):
        if self._data.get(name, None) is None:
            self._data[name] = []

        self._data[name].extend(values)

    def get_value(
            self,
            name: str,
            default: Optional[str] = None) -> Optional[str]:
        values = self._data.get(name, None)

        self._read[name] = True

        if values is None or len(values) == 0:
            return default
        elif len(values) > 1:
            raise Exception(f'Too much values: {name}')

        return values[0]

    def get_list(self, name: str) -> List[str]:
        values = self._data.get(name, None)

        self._read[name] = True

        if values is None:
            return []

        return values



from stx.design.components import Component
from stx.design.document import Document

from ._parsing import parse_component
from ._source import Source


def parse_from_file(document: Document, file_path: str) -> Component:
    with open(file_path, 'r') as stream:
        source = Source(stream, file_path)

        return parse_component(document, source)


from __future__ import annotations

from typing import List

from stx.design.components import Heading


class IndexNode:

    def __init__(self, heading: Heading):
        self.heading = heading
        self.level = heading.level
        self.nodes: List[IndexNode] = []

    def add(self, node: IndexNode):
        self.nodes.append(node)



from __future__ import annotations

import re
from io import StringIO
from typing import List, Iterable, Optional, TextIO

from stx.design.attributes_map import AttributesMap

from stx.utils.strs import crop_text


class Component:
    _attributes = None

    @property
    def attributes(self) -> AttributesMap:
        if self._attributes is None:
            self._attributes = AttributesMap()
        return self._attributes

    def get_text(self) -> str:
        output = StringIO()

        self.write_text(output)

        return output.getvalue()

    def walk(self) -> Iterable[Component]:
        for child in self.get_children():
            yield child

            for c in child.walk():
                yield c

    def write_text(self, output: TextIO):
        raise NotImplementedError()

    def get_children(self) -> List[Component]:
        raise NotImplementedError()

    def __repr__(self):
        raise NotImplementedError()


class Composite(Component):

    def __init__(self, components: List[Component]):
        self.components = components

    def __repr__(self):
        return f'Composite<{len(self.components)} component(s)>'

    def write_text(self, output: TextIO):
        for component in self.components:
            component.write_text(output)

    def get_children(self) -> List[Component]:
        return self.components


class Heading(Component):

    def __init__(self, content: Component, level: int):
        self.content = content
        self.level = level
        self.number = None

    def __repr__(self):
        return f'Heading<{crop_text(self.get_text(), 10)}>'

    def write_text(self, output: TextIO):
        self.content.write_text(output)

    def get_children(self) -> List[Component]:
        return [self.content]


class RawText(Component):

    def __init__(self, text: str):
        self.text = text

    def __repr__(self):
        return f'RawText<{crop_text(self.text, 10)}>'

    def write_text(self, output: TextIO):
        output.write(self.text)

    def get_children(self) -> List[Component]:
        return []


class CodeBlock(Component):

    def __init__(self, code: str, flavor: str):
        self.code = code
        self.flavor = flavor

    def __repr__(self):
        return f'CodeBlock<{crop_text(self.code, 10)}>'

    def write_text(self, output: TextIO):
        output.write(self.code)

    def get_children(self) -> List[Component]:
        return []


class TableRow(Component):

    def __init__(self, cells: List[Component], header: bool):
        self.cells = cells
        self.header = header

    def __repr__(self):
        return f'TableRow<{len(self.cells)} cell(s)>'

    def write_text(self, output: TextIO):
        for cell in self.cells:
            cell.write_text(output)

    def get_children(self) -> List[Component]:
        return self.cells


class Table(Component):

    def __init__(self, rows: List[TableRow]):
        self.rows = rows
        self.caption = None
        self.number = None

    def __repr__(self):
        return f'Table<{len(self.rows)} row(s)>'

    def write_text(self, output: TextIO):
        for row in self.rows:
            row.write_text(output)

    def get_children(self) -> List[Component]:
        if self.caption is not None:
            return [self.caption, *self.rows]

        return self.rows


class Figure(Component):

    def __init__(self, content: Component, caption: Component):
        self.content = content
        self.caption = caption
        self.number = None

    def __repr__(self):
        return f'Figure<{crop_text(self.caption.get_text(), 10)}>'

    def write_text(self, output: TextIO):
        self.caption.write_text(output)
        self.content.write_text(output)

    def get_children(self) -> List[Component]:
        return [self.caption, self.content]


class ListBlock(Component):

    def __init__(self, items: List[Component], ordered: bool):
        self.items = items
        self.ordered = ordered

    def __repr__(self):
        return f'ListBlock<{len(self.items)} item(s)>'

    def write_text(self, output: TextIO):
        for item in self.items:
            item.write_text(output)

    def get_children(self) -> List[Component]:
        return self.items


class StyledText(Component):

    def __init__(self, contents: List[Component], style: str):
        self.contents = contents
        self.style = style

    def __repr__(self):
        return f'StyledText<{crop_text(self.get_text(), 10)}>'

    def write_text(self, output: TextIO):
        for content in self.contents:
            content.write_text(output)

    def get_children(self) -> List[Component]:
        return self.contents


class LinkText(Component):

    def __init__(self, contents: List[Component], reference: Optional[str]):
        self.contents = contents
        self.reference = reference

    def __repr__(self):
        return f'LinkText<{crop_text(self.get_text(), 10)}>'

    def write_text(self, output: TextIO):
        for content in self.contents:
            content.write_text(output)

    def get_children(self) -> List[Component]:
        return self.contents

    @property
    def internal(self) -> bool:
        # TODO refactor to external
        return self.reference is not None and not re.match(r'(?i)^([a-z]+:)?//', self.reference)


class PlainText(Component):

    def __init__(self, text: str):
        self.text = text

    def __repr__(self):
        return f'PlainText<{crop_text(self.text, 10)}>'

    def write_text(self, output: TextIO):
        output.write(self.text)

    def get_children(self) -> List[Component]:
        return []


class TextBlock(Component):

    def __init__(self, components: List[Component]):
        self.components = components

    def __repr__(self):
        return f'TextBlock<{crop_text(self.get_text(), 10)}>'

    def write_text(self, output: TextIO):
        for component in self.components:
            component.write_text(output)

    def get_children(self) -> List[Component]:
        return self.components



from typing import Optional


def generate_error_message(message, file_path, line_index, column_index):
    location = ''

    if file_path:
        location += file_path

    if line_index is not None and column_index is not None:
        location += f' @ Line {line_index + 1}, Column {column_index + 1}'
    elif line_index is not None:
        location += f' @ Line {line_index + 1}'
    elif column_index is not None:
        location += f' @ Column {column_index + 1}'

    return f'{location}:\n{message}'


class ParseError(Exception):

    def __init__(
            self,
            message: str,
            file_path: Optional[str] = None,
            line_index: Optional[int] = None,
            column_index: Optional[int] = None):
        super().__init__(generate_error_message(message, file_path, line_index, column_index))




from __future__ import annotations

from typing import List

from stx import logger
from stx.design.index_node import IndexNode
from stx.utils.stack import Stack
from stx.design.components import Figure, Table, Heading, Component


def build_numbering(document: Component) -> List[IndexNode]:
    logger.info('Building numbering...')

    index: List[IndexNode] = []
    stack: Stack[IndexNode] = Stack()
    figures: List[Figure] = []
    tables: List[Table] = []

    for content in document.walk():
        if isinstance(content, Heading):
            node = IndexNode(content)
            level = content.level

            if stack.empty():
                index.append(node)
                stack.push(node)
            elif level > stack.peek().level:
                stack.peek().add(node)
                stack.push(node)
            elif level == stack.peek().level:
                if stack.empty():
                    raise Exception('Invalid state')
                stack.pop()
                if stack.empty():
                    index.append(node)
                else:
                    stack.peek().add(node)
                stack.push(node)
            else:
                while not stack.empty() and level <= stack.peek().level:
                    stack.pop()

                if stack.empty():
                    index.append(node)
                else:
                    stack.peek().add(node)

                stack.push(node)
        elif isinstance(content, Table):
            tables.append(content)
        elif isinstance(content, Figure):
            figures.append(content)

    logger.info('Applying numbering...')
    apply_heading_numbers('', index, 3)
    apply_table_numbers(tables)
    apply_figure_numbers(figures)

    return index


def apply_heading_numbers(
        parent: str, nodes: List[IndexNode], max_levels: int, level: int = 1):
    count = 1
    for node in nodes:
        number = f'{count}.'

        if parent:
            number = f'{parent}{number}'

        node.heading.number = number

        count += 1

        if level < max_levels:
            apply_heading_numbers(number, node.nodes, max_levels, level + 1)


def apply_table_numbers(tables: List[Table]):
    count = 1

    for table in tables:
        table.number = f'Table {count}.'
        count += 1


def apply_figure_numbers(figures: List[Figure]):
    count = 1

    for figure in figures:
        figure.number = f'Figure {count}.'
        count += 1



from stx.rendering.html5.renderer import render_document
from stx.loading.loaders import from_file
from stx.rendering.html5.writer import HtmlWriter


def main():
    file_path = '/Users/sergio/bm/docs/src/index.stx'

    document = from_file(file_path)

    with open("/Users/sergio/bm/docs/docs/index.html", "w", encoding="utf-8") as f:
        render_document(document, HtmlWriter(f))

    print(document)


if __name__ == '__main__':
    main()



heading6_mark = '======'
heading5_mark = '====='
heading4_mark = '===='
heading3_mark = '==='
heading2_mark = '=='
heading1_mark = '='
unordered_list_item_mark = '-'
ordered_list_item_mark = '.'
table_h_row_mark = '|='
table_d_row_mark = '|-'
table_cell_mark = '|'
pre_caption_mark = '>'
post_caption_mark = '<'
content_box_mark = '%%%'
code_block_mark = '```'
comment_block_mark = '///'
attribute_mark = '@'
directive_mark = '#'

escape_mark = '\\'

# The order matters for finding the level
heading_marks = [
    heading1_mark,
    heading2_mark,
    heading3_mark,
    heading4_mark,
    heading5_mark,
    heading6_mark,
]

flat_block_marks = [
    content_box_mark,
    code_block_mark,
    comment_block_mark,
]

# The order matters for parsing
reserved_block_marks = [
    heading6_mark,
    heading5_mark,
    heading4_mark,
    heading3_mark,
    heading2_mark,
    heading1_mark,
    unordered_list_item_mark,
    ordered_list_item_mark,
    table_d_row_mark,
    table_h_row_mark,
    table_cell_mark,
    pre_caption_mark,
    post_caption_mark,
    attribute_mark,
    directive_mark,
    code_block_mark,
    content_box_mark,
    comment_block_mark,
]

strong_mark = '*'
emphasis_mark = '_'
code_mark = '`'
begin_link_mark = '['
begin_macro_mark = '<'

reserved_text_marks = [
    strong_mark,
    emphasis_mark,
    code_mark,
    begin_link_mark,
    begin_macro_mark,
]


def get_heading_level(mark):
    if mark not in heading_marks:
        raise Exception(f'`{mark}` is not a heading mark.')
    return heading_marks.index(mark) + 1



from __future__ import annotations

import os
import string
from typing import Optional, List, Tuple

from stx.utils.stack import Stack
from stx.design.components import Composite, Component, TextBlock, Heading, \
    RawText, CodeBlock, Table, Figure, ListBlock, TableRow, StyledText, \
    LinkText, PlainText
from stx.design.document import Document
from stx.utils.files import resolve_sibling, walk_files

from ._source import Source
from ._tape import Tape
from ._composer import Composer
from . import _marks


def parse_styled_text(tape: Tape, stop_marks: Stack) -> Optional[StyledText]:
    if tape.test('*'):
        style = 'strong'
    elif tape.test('_'):
        style = 'emphasized'
    elif tape.test('`'):
        style = 'code'
    else:
        return None

    delimiter = tape.read()

    stop_marks.push(delimiter)

    contents = parse_inline_text(tape, stop_marks)

    stop_marks.pop()

    if not tape.pull(delimiter):
        raise Exception('styled text not closed')

    return StyledText(contents, style)


def parse_link_text(tape: Tape, stop_marks: Stack) -> Optional[LinkText]:
    if not tape.pull('['):
        return None

    stop_marks.push(']')

    contents = parse_inline_text(tape, stop_marks)

    stop_marks.pop()

    if not tape.pull(']'):
        raise Exception('link not closed')

    if tape.pull('('):
        reference = ''

        while not tape.test(')'):
            reference += tape.read()

        if not tape.pull(')'):
            raise Exception('reference not closed')
    else:
        reference = None

    return LinkText(contents, reference)


def parse_plain_text(tape: Tape, stop_marks: Stack) -> Optional[PlainText]:
    content = ''

    while tape.alive():
        if tape.test(stop_marks.peek()):
            break
        elif tape.any('*', '_', '`', '['):
            break
        elif tape.pull('\\'):
            if tape.test(stop_marks.peek()):
                content += tape.read()
            elif tape.any('*', '_', '`', '[', ']', '(', ')', '\\'):
                content += tape.read()
            else:
                content += '\\'
        else:
            content += tape.read()

    if len(content) == 0:
        return None

    return PlainText(content)


def parse_inline_text(tape: Tape, stop_marks: Stack) -> List[Component]:
    contents = []

    while tape.alive():
        if tape.test(stop_marks.peek()):
            break

        content = (parse_styled_text(tape, stop_marks)
                   or parse_link_text(tape, stop_marks)
                   or parse_plain_text(tape, stop_marks))

        if content is not None:
            contents.append(content)

    return contents


def parse_text_block(document: Document, source: Source, composer: Composer, stop_mark: Optional[str]):
    stop_marks = Stack()

    if stop_mark is not None:
        stop_marks.push(stop_mark)

    components = parse_inline_text(Tape(source), stop_marks)

    # TODO

    composer.push(TextBlock(components))


def parse_heading_block(document: Document, source: Source, level: int, composer: Composer):
    content = parse_component(document, source)

    composer.push(Heading(content, level))


def process_include(document: Document, source: Source, include_paths: list, composer: Composer):
    for include_path in include_paths:
        target_path = resolve_sibling(source.file_path, include_path)
        file_paths = []

        if os.path.isdir(target_path):
            for file_path in walk_files(target_path):
                file_paths.append(file_path)
        else:
            file_paths.append(target_path)

        for file_path in sorted(file_paths):
            print(f'Processing: {file_path}...')
            if file_path.endswith('.stx'):
                with open(file_path, 'r') as stream:
                    source = Source(stream, file_path)

                    component = parse_component(document, source)

                    composer.push(component)
            else:
                # TODO add support for more extensions
                with open(file_path, 'r', encoding='UTF-8') as f:
                    content = f.read()

                composer.push(RawText(content))


def parse_name_values(tape: Tape) -> Tuple[str, List[str]]:
    tape.skip(' ')

    if not tape.test(string.ascii_letters):
        raise tape.error('Expected to read a name.')

    name = tape.read()

    while tape.test(string.ascii_letters + string.digits + '._-'):
        name += tape.read()

    tape.skip(' ')

    values = []

    if tape.pull('['):
        tape.skip(' ')

        value = ''

        while tape.alive():
            if tape.test(']'):
                break
            elif tape.pull(','):
                values.append(value)
                value = ''

                tape.skip(' ')
            elif tape.pull('\\'):
                c = tape.read()

                if c not in [',', ']']:
                    value += '\\'

                value += c
            else:
                value += tape.read()

        if not tape.pull(']'):
            raise tape.error('Expected to read a `]`')

        if value:
            values.append(value)

    tape.skip(' \n')

    if tape.alive():
        raise tape.error('Unexpected content')

    return name, values


def process_link(document: Document, source: Source, values: list, composer: Composer):
    if len(values) < 2:
        raise Exception('Expected rel and files')

    rel = values[0]

    document.links.add_values(rel, values[1:])


def parse_directive(document: Document, source: Source, composer: Composer):
    tape = Tape(source)

    name, values = parse_name_values(tape)

    if name == 'link':
        process_link(document, source, values, composer)
    elif name == 'include':
        process_include(document, source, values, composer)
    else:
        raise Exception(f'Not implemented directive: {name}')


def parse_attribute(document: Document, source: Source, composer: Composer):
    tape = Tape(source)

    name, values = parse_name_values(tape)

    composer.push_attribute(name, values)


def parse_code_block(document: Document, source: Source, composer: Composer):
    line = source.pop_line()

    flavor = line.get_current_text()

    lines = []

    begin_index = source.peek_begin_index()

    while True:
        line = source.pop_line()

        if line.test_current_text(_marks.code_block_mark):
            break

        lines.append(line.raw_text[begin_index:])

    composer.push(CodeBlock('\n'.join(lines), flavor))


def parse_pre_caption(document: Document, source: Source, composer: Composer):
    composer.pending_caption = parse_component(document, source)


def parse_post_caption(document: Document, source: Source, composer: Composer):
    caption = parse_component(document, source)

    last = composer.get_last_content()

    if isinstance(last, Table):
        last.caption = caption
    else:
        content = composer.pop()

        figure = Figure(content, caption)

        composer.push(figure)


def parse_list_item(document: Document, source: Source, composer: Composer, ordered: bool):
    last = composer.get_last_content()

    if isinstance(last, Figure):
        last = last.content

    if isinstance(last, ListBlock) and last.ordered == ordered:
        active_list = last
    else:
        active_list = ListBlock([], ordered)

        composer.push(active_list)

    content = parse_component(document, source)

    active_list.items.append(content)


def parse_table_row(document: Document, source: Source, composer: Composer, header: bool):
    last = composer.get_last_content()

    if isinstance(last, Table):
        active_table = last
    else:
        active_table = Table([])

        composer.push(active_table)

    cells = []

    stop_mark = _marks.table_cell_mark

    while True:
        cell = parse_component(document, source, stop_mark)

        cells.append(cell)

        line = source.try_line()

        if line is None:
            break
        elif not line.get_current_text().startswith(stop_mark):
            break

        line.indentation += len(stop_mark)

    active_table.rows.append(TableRow(cells, header))


def parse_component(document: Document, source: Source, stop_mark: Optional[str] = None) -> Component:
    composer = Composer()

    while source.is_alive():
        line = source.get_line()

        if stop_mark and line.get_current_text().startswith(stop_mark):
            break

        block_mark = line.consume_block_info()

        if block_mark is None:
            parse_text_block(document, source, composer, stop_mark)
        else:
            is_flat = block_mark in _marks.flat_block_marks

            if not is_flat:
                source.push_begin_index(line.indentation)

            if block_mark in _marks.heading_marks:
                parse_heading_block(document, source, _marks.get_heading_level(block_mark), composer)
            elif block_mark == _marks.directive_mark:
                parse_directive(document, source, composer)
            elif block_mark == _marks.attribute_mark:
                parse_attribute(document, source, composer)
            elif block_mark == _marks.code_block_mark:
                parse_code_block(document, source, composer)
            elif block_mark == _marks.pre_caption_mark:
                parse_pre_caption(document, source, composer)
            elif block_mark == _marks.post_caption_mark:
                parse_post_caption(document, source, composer)
            elif block_mark == _marks.unordered_list_item_mark:
                parse_list_item(document, source, composer, False)
            elif block_mark == _marks.ordered_list_item_mark:
                parse_list_item(document, source, composer, True)
            elif block_mark == _marks.table_d_row_mark:
                parse_table_row(document, source, composer, False)
            elif block_mark == _marks.table_h_row_mark:
                parse_table_row(document, source, composer, False)
            else:
                raise source.error(f'Not implemented block mark: {block_mark}')

            if not is_flat:
                source.pop_begin_index()

    components = composer.compile()

    if len(components) == 1:
        return components[0]

    return Composite(components)



import string
from typing import Dict, Optional, List

from stx.design.components import Component


def make_ref(base: str, length_hint=40) -> str:
    result = []

    for c in base.lower():
        if c in string.ascii_lowercase or c in string.digits:
            result.append(c)
        elif len(result) >= length_hint:
            break
        elif len(result) == 0 or result[-1] != '-':
            result.append('-')

    return ''.join(result)


def compute_ref(content: Component, count: Optional[int]) -> str:
    plain_text = content.get_text().strip()

    if count is not None:
        plain_text += f'-{count}'

    return make_ref(plain_text)


class RefMap:

    def __init__(self):
        self._ref_cons: Dict[str, Component] = {}
        self._con_refs: Dict[Component, List[str]] = {}

    def contains_ref(self, ref: str):
        return ref in self._ref_cons.keys()

    def register_content(self, content: Component, main=False) -> str:
        count = 0

        while True:
            ref = compute_ref(content, count if count > 0 else None)

            if not self.contains_ref(ref):
                break

            count += 1

        self.register_ref(ref, content, main)

        return ref

    def register_ref(self, ref: str, content: Component, main=False):
        if self.contains_ref(ref):
            raise Exception(f'Link already registered: {ref}')

        self._ref_cons[ref] = content

        if self._con_refs.get(content, None) is None:
            self._con_refs[content] = []

        if ref not in self._con_refs[content]:
            if main:
                self._con_refs[content].insert(0, ref)
            else:
                self._con_refs[content].append(ref)

    def get_refs(self, content: Component) -> List[str]:
        return self._con_refs.get(content, [])

    def get_main_ref(self, content: Component) -> Optional[str]:
        refs = self.get_refs(content)

        if len(refs) == 0:
            return None

        return refs[0]

    def get_other_refs(self, content: Component) -> List[str]:
        refs = self.get_refs(content)

        if len(refs) <= 1:
            return []

        return refs[1:]

    def get_content(self, ref: str) -> Optional[Component]:
        return self._ref_cons.get(ref, None)
